import os
import shelve
from datetime import datetime
from zipfile import ZIP_DEFLATED, ZipFile

import Containers
import Folders
from functions import format_time
from Generators import Generators
from Localization import Localization
from LocalizationCache import LocalizationCache


class ModBase:

    '''
    some universal framework that I seem
    to be duplicating a lot as I generate mini mods.
        Conventions:
            generate() isn't mandatory, but is a good idea.
            subclasses don't require their own masks for gen_mod

    notes:
        I need a better way to handle localizations.
        maybe data files?

        better block generators...
            x.y='yes'
            x.b = block()
            x.b.a='no'
                perhaps

        add a readme to the mod-dir...
            e.g. something like
                This minimod was automatically generated by script.
    '''

    def __init__(self):
        # standard folder definitions; nothing is generated if empty.
        self.localizations = Folders.LocalizationFolder()
        self.decisions = Folders.DecisionFolder()
        self.events = Folders.EventFolder()
        self.event_modifiers = Folders.ModFolder('common\\event_modifiers\\',
                                         Containers.ContainerBase, 'mods.txt')
        self.buildings = Folders.BuildingFolder()
        self.traits = Folders.ModFolder('common\\traits\\',
                                Containers.ContainerBase, 'traits.txt')
        self.titles = Folders.ModFolder('common\\landed_titles\\',
                                Containers.ContainerBase, 'titles.txt')
        self.interface = Folders.ModFolder('interface\\',
                                   Containers.ContainerBase, '.gfx')
        self.on_actions = Folders.ModFolder('common\\on_actions\\',
                                    Containers.ContainerBase, 'on_actions.txt')
        self.scripted_triggers = Folders.ModFolder('common\\scripted_triggers\\',
                                           Containers.ScriptedObjectContainer, 'scripted_triggers.txt')
        self.scripted_effects = Folders.ModFolder('common\\scripted_effects\\',
                                          Containers.ScriptedObjectContainer, 'scripted_effects.txt')
        self.readme_folder = Folders.ModFolder('', Containers.ContainerBase, 'readme.txt')

        self.folders = [self.localizations,
                        self.decisions,
                        self.events,
                        self.event_modifiers,
                        self.buildings,
                        self.traits,
                        self.titles,
                        self.interface,
                        self.on_actions,
                        self.scripted_triggers,
                        self.scripted_effects, ]
        self.submods = []
        self._goslate = None
        # translation cache
        self._gscache = None

        self.generators=Generators(self)

        # these masks might be overwritten by subclassing mods;
        #    if create_mod is called....if
        #    its just stored as an additional storage place for events, there
        #    is no need to customize anything beyond filename_mask
        self.filename_mask = 'subclasses_shouldhave_customized_this_{}_{}'
        self.mod_folder_mask = 'modbase_folder_mask_{}'
        self.mod_file_mask = 'modbase_file_mask_{}.mod'
        self.vanilla_mod_name = 'modbase vanilla mod'
        self.submod_mod_name = '{} (modbase SubMod)'
        self.gs_cache_name = self.filename_mask.format(
            'translation_cache', '.shelf')

        self.readmes = {}
        self._last_clear_ts = None
        self._initialized_ts = datetime.now()
        self._written=0#keeps track of number of files written

    def add_folder(self, folder_name, folder_path, default_file='default.txt'):
        setattr(self, folder_name, Folders.ModFolder(
            folder_path, Containers.ContainerBase, default_file))
        self.folders.append(getattr(self, folder_name))

    def add_readme(self, name, text):
        if name not in self.readmes:
            self.readmes[name] = text
        else:
            self.readmes[name] += text

    def get_readmes(self):
        ret = []
        for readme in self.readmes:
            ret.append((readme, self.readmes[readme]))
        for mod in self.submods:
            ret = ret + mod.get_readmes()
        return ret

    def _handle_misc(self):
        # first, append any info in submods to my info
        merged = {}
        for name, string in self.get_readmes():
            if name not in merged:
                merged[name] = string
            else:
                merged[name] = merged[name] + string
        for name in merged:
            self.readme_folder.add_preserialized(merged[name], name)

    def _open_cache(self):
        if not self._gscache:
            dest = os.path.join(os.getcwd(), 'data')
            if not os.path.exists(dest):
                os.makedirs(dest)
            shelf_file = os.path.join(dest, self.gs_cache_name)
            # print shelf_file
            # self._gscache = shelve.open(self.gs_cache_name, 'c',
            # writeback=True)
            self._gscache = LocalizationCache(shelf_file)
            self._gscache.open()

    def _cache(self, name, base, dest, result):
        if not self._gscache:
            self._open_cache()
        # base is always english
        # if for some reason I changed the localization string, I should
        # probably reacquire it.
        if self._gscache.get(name, 'en') != base:
            self._gscache.set(name, 'en', base)
        self._gscache.set(name, dest, result)

    def _retrieve_from_cache(self, name, base, dest):
        # base is always english
        if not self._gscache:
            self._open_cache()
        stored = self._gscache.get(name, dest)
        if not stored:
            self._cache(name, base, dest, ' ')
            return self._gscache.get(name, dest)
        # if base in self._gscache:
            # if dest in self._gscache[base]:
            # return self._gscache[base][dest]
        # return None

    def append_to_readme(self, readme_name, text):
        if not self.readmes.has_key(readme_name):
            self.readmes[readme_name] = text
        else:
            self.readmes[readme_name] = self.readmes[report_filename] + text

    def elapsed_since_clear(self):
        return datetime.now() - self._last_clear_ts

    def elapsed_since_init(self):
        return datetime.now() - self._initialized_ts

    def initialize(self):
        #this should probably be overwritten by subclasses
        #and is intended to initialize subclasses.
        return

    def generate(self):
        # this can be overriden by subclasses,
        #particularly if they odn't want to use generators.
        self.initialize()
        self.generators.generate()
        self.generators.localize()
        return

    def add_localization(self, name, english,
                         french='',
                         german='',
                         spanish='', filename=None):
        self.add_loc(name, english, french, german, spanish, filename)

    def add_loc(self, name, english,
                french='',
                german='',
                spanish='',
                filename=None):
        '''
        A better way to handle localization:
            register the localization strings that are necessary
            and generate a datafile; which is then filled out
            by hand.

            At least as an option...maybe a 'load_localizations'
            command to parse in a local datafile.

        use goslate (if installed) to populate that datafile?
        '''
        loc = Localization(name)
        loc.english = english
        if not french:
            french = self._gstrans(name, english, 'fr')
        loc.french = french
        if not german:
            german = self._gstrans(name, english, 'de')
        loc.german = german
        if not spanish:
            spanish = self._gstrans(name, english, 'es')
        loc.spanish = spanish
        self.localizations.add(loc, filename)
        # self.localizations[name] = loc

    def _gstrans(self, name, base, target):
        old = self._retrieve_from_cache(name, base, target)
        if old:
            old = old.strip()  # strip out extra whitespace.
        return old
        # if I find a Goslate replacement, the below becomes relevant again.
        # if old:
        # return old
        # if not self._goslate:
        #import goslate
        #self._goslate = goslate.Goslate()
        #new = self._goslate.translate(base, target)
        #self._cache(name, base, target, new)
        # return new

    def _shorten_mod_name(self, dirname):
        return dirname.replace(' ', '_')

    # def _generate_wrapper(self):
        # time generate as a whole?
        # self.generate()

    def things_written(self):
        count=0
        for mod in self.submods:
            count+=mod.things_written()
        count+=self._written
        return count
    
    def _report_writetimes(self,starttime):
        mask = 'Writing {} files took {}.\n'
        time = format_time(datetime.now() - starttime)  
        elapsed=mask.format(self.things_written(),time)
        self.add_readme('script_output.txt',elapsed)

    def write_zip(self, azip, mod_name):
        start=datetime.now()
        self._written=0
        short = self._shorten_mod_name(mod_name)
        # lets do submods first, since they might write to our readme or
        # something silly.
        for mod in self.submods:
            mod.write_zip(azip, mod_name)
        self.generate()
        for afolder in self.folders:
            afolder.write_zip(azip, short, self.filename_mask)
            self._written+=afolder.things_written()
        # do readme at the end.
        self._report_writetimes(start)
        self._handle_misc()
        self.readme_folder.write_zip(azip, short, self.filename_mask)

    def write(self, folder, mod_name):
        start=datetime.now()
        self._written=0
        short = self._shorten_mod_name(mod_name)
        # lets do submods first, since they might write to our readme or
        # something silly.
        for mod in self.submods:
            mod.write(folder, mod_name)
        self.generate()
        for afolder in self.folders:
            afolder.write(folder, short, self.filename_mask)
            self._written+=afolder.things_written()
        # do readme at the end.
        self._report_writetimes(start)
        self._handle_misc()
        self.readme_folder.write(folder, short, self.filename_mask)

    def clear(self):
        self.last_clear = datetime.now()
        for folder in self.folders:
            folder.clear()

    def zipname(self, short):
        return self.mod_file_mask.format(short).replace('.mod', '.zip')

    def _gen_mod_file(self, folder, mod_name, mod_folder, short, zipfile=False):
        # print folder
        # print mod_name
        if 'VANILLA' not in mod_name:
            shortname = mod_name[:10]
            if len(mod_name) > 10:
                shortname = shortname + '...'
            out = 'name = "' + self.submod_mod_name.format(shortname) + '"\n'
            depstr = 'dependencies = { "' + mod_name + '" }'
        else:
            out = 'name = "{}"\n'.format(self.vanilla_mod_name)
            depstr = ''
        if not zipfile:
            out += 'path = "mod/' + mod_folder + '"\n'
        else:
            out += 'path = "mod/' + self.zipname(short) + '"\n'
        out += depstr
        mod_name = folder + '\\' + self.mod_file_mask.format(short)
        with open(mod_name, 'w') as _fp:
            _fp.write(out)

    def create_mod(self, destination, mod_name, zipfile=False):
        short = mod_name.replace(' ', '_').replace(
            '-', '').replace(':', '_').strip()
        mod_folder = self.mod_folder_mask.format(short)
        if not zipfile:
            self.write(os.path.join(destination, mod_folder), short)
        else:
            azipfile = destination + '\\' + self.zipname(short)
            print azipfile
            with ZipFile(azipfile, 'w', ZIP_DEFLATED) as azip:
                self.write_zip(azip, mod_name)
        self._gen_mod_file(destination, mod_name, mod_folder, short, zipfile)
        if self._gscache:
            if self._gscache.changes > 0:
                # ensure language cache changes are written out.
                self._gscache.write()
